#ifndef CursorWindowViewportAdjust_included_p
#define CursorWindowViewportAdjust_included_p

///@file
///  This file defines the template class, ViewportAdjust, which exists to allow you to resize 
///  cursor window viewports (ie rectangular screen subregions).  The code is written maximize
///  generality, so it may be directly usable on other containers of rectangles.
///

#include <vector>
#include <set>
#include <map>
#include <utility> 
#include <cxxtls/foreach.h>
#include <string>

namespace cxxtls
{

template<class Config>
struct ViewportAdjust
    ///  
    /// Template struct ViewportAdjust exists to implement the increasing in size of rectangular, 
    /// cursor style windows which are part of tiled window "screen" set.  The resizing occurs
    /// within a fixed total "screen" size.  This class is desigend to resize one window, within
    /// the existing screen limits, at the expense of other windows which are reduced in size.  
    /// Here is an example of of a tiled window set:
    ///
    ///@code
    ///
    ///   +/home/user/dir-----+/fred/bill---------+
    ///   | files in home     | contents of bill  |
    ///   | more files        +/hank--------------+
    ///   +/other-------------+  contents of      |
    ///   |  other files      |  file /hank       |
    ///   +-------------------+-------------------+
    ///
    ///@endcode
    ///
    /// In this case, there are 4 open viewers, on 4 different files/directories.  They are not
    /// all the same size on the screen.  Note that any of the above 4 sub-windows could themselves
    /// be split.  There is a minimum limit to the height and width of viewers (4, 20).
    ///
    /// Member function widenViewport(), is called so as to increase either the width or the
    /// height of one member of the input viewport set (ie sub-window set).  The Config template
    /// parameter defines what happens.  See the Config instances in file @ref viewermanager.cxx
    /// for examples.
    ///
    /// The class is comprised of a bunch of static methods that support the primary 
    /// interface, widenViewport(), which lets you specify a container of viewport pointers 
    /// and a specific viewport in that container to be widened.  The function will resize 
    /// the named viewport based on template parameters -- either wider or taller.  Other 
    /// windows may or may not be moved based on the requirements of re-adjusting the 
    /// specified window.
    ///
    /// The viewports (ie subwindows) in the viewport container have an "origin" and a "size".
    /// The origin is a ScreenLocation that defines the upper left hand corner of the
    /// viewport.  The size begins in that "cursor cell" and defines the total number of
    /// cells in the horizontal and vertical dimensions.  Both the origin and size are of
    /// type ScreenLocation, which must be defined in the Config template parameter as a
    /// nested typedef -- as must class ViewPort.
    ///
    /// Template parameter class, Config, defines the following
    /// @li nested typedef ScreenLocation -- represents and x/y pair or row/col pair
    /// @li nested typedef ViewPort -- defines the object being moved around by this adjustment
    /// @li nested typedef Manager -- defines the owner of the set of all viewports -- only used for debug
    /// @li various accessor methods, described below:
    ///
    /// The accessor methods are as follows:
    ///
    ///@code
    ///  static void setPrimaryDimension(ScreenLocation &p, int value);       // set row or col in location
    ///  static int getPrimaryDimension(ScreenLocation const &p);             // get row or col from location
    ///  static void setSecondaryDimension(ScreenLocation &p, int value);     // set col or row in location
    ///  static int getSecondaryDimension(ScreenLocation const &p);           // get col or row from location
    ///  static ScreenLocation  getOrigin(ViewPort const *vp);                // get upper left corner of viewport
    ///  static ScreenLocation  getSize(ViewPort const *vp);                  // get width and height of viewport
    ///  static void move(ViewPort *vp, ScreenLocation const &origin, ScreenLocation const &size); // move viewport
    ///@endcode
    /// 
{
     typedef typename Config::ScreenLocation  ScreenLocation;   ///<  An x,y pair or row,col pair, depending on Config

     typedef typename Config::ViewPort        ViewPort;         ///<  A sub-window.  The caller of windowViewport() 
                                                                ///   passes in a collection of sub-windows and asks for
                                                                ///   one of them to be widened based on member functions
                                                                ///   in the Config template parameter.

     typedef typename Config::Manager         Manager;          ///<  A helper class that has the viewerFullname() method
                                                                ///   which is used ONLY for debug, so it can return
                                                                ///   empty string if you don't want to bother with it.

     static void setPrimaryDimension(ScreenLocation &p, int value)
         ///  Given a "screen location", set the primary dimension of that location
         ///  using the helper function, Config::setPrimaryDimension() which is
         ///  provided as part of the template parameter, Config, for this 
         ///  invocation.  The primary dimension will be either row or column, depending
         ///  on Config.
     {
         // set either the row or the column based on functions 
         // defined in the Config parameter

         Config::setPrimaryDimension(p, value);
     }

     static int getPrimaryDimension(ScreenLocation const &p)
         ///  Given a screen location, get the primary dimension value.  This will be
         ///  either the row or the column of the screen location depending on the
         ///  implementation of the Config::getPrimaryDimension() method provided
         ///  as part of the template signature.
     {
         return Config::getPrimaryDimension(p);
     }

     static void setSecondaryDimension(ScreenLocation &p, int value)
         ///  Given a "screen location", set the secondary dimension of that location
         ///  using the helper function, Config::setSecondaryDimension() which is
         ///  provided as part of the template parameter, Config, for this 
         ///  invocation.  The secondary dimension will be either column or row, depending
         ///  on Config.
     {
         Config::setSecondaryDimension(p, value);
     }

     static int getSecondaryDimension(ScreenLocation const &p)
         ///  Give a "screen location", return the value of the non-primary dimension as
         ///  defined by Config::getSecondaryDimension() -- which is defined by the
         ///  template parameter.
     {
         return Config::getSecondaryDimension(p);
     }

     static ScreenLocation  getOrigin(ViewPort const *vp)
         ///  Given a subwindow, return the screen location of the upper left corner.
     {
        return Config::getOrigin(vp);
     }

     static ScreenLocation  getSize(ViewPort const *vp)
         ///  Given a sub-window, return the width and height as a ScreenLocation.
     {
        return Config::getSize(vp);
     }

     static void move(ViewPort *vp, ScreenLocation const &origin, ScreenLocation const &size)
         ///  Upon completion of the tile re-layout process, the function, widenViewport() will
         ///  call the move() method on any sub-windows affected by its operation.
     {
        Config::move(vp, origin, size);
     }

     struct Box
         /// a rectangle and members to perform various tests against it.
         ///
         /// instead of x,y or row,col, we have primaryDimension,secondaryDimension.  Its the same thing
         /// just named funny.  This is done so that the accessors used in the rest of the algorithms
         /// in the class will work on Box.
         ///
         /// The box is defined by two points:  the upper left corner's ScreenLocation, and the lower right
         /// corner's ScreenLocation.  However, the lower right corner is not "in" the box.  It is one row
         /// and one column outside it.
         ///
         /// Note the words "upper", "left", "right", and "lower", are figurative.  The actually values are
         /// based on primary and secondary dimensions.  The least value of the primary dimension and the
         /// secondary dimension are called "upper" and "left".  In fact the row and columns may be swapped
         /// conceptually.
     {
         ScreenLocation upperLeft_;    // actually in the box
         ScreenLocation lowerRight_;   // outside the box

         Box(ScreenLocation const &least, ScreenLocation const &size)
             ///  Construct a box given a location and dimensions.
         {
             upperLeft_ = least;

             setPrimaryDimension  ( lowerRight_, getPrimaryDimension(least)   + getPrimaryDimension(size)  );
             setSecondaryDimension( lowerRight_, getSecondaryDimension(least) + getSecondaryDimension(size) );
         }

         Box(ViewPort const *vp)
             // Construct a box given a pointer to a sub-window.
         {
             ScreenLocation origin = getOrigin(vp);
             ScreenLocation size   = getSize(vp);

             upperLeft_ = origin;

             setPrimaryDimension  ( lowerRight_, getPrimaryDimension(origin)   + getPrimaryDimension(size)  );
             setSecondaryDimension( lowerRight_, getSecondaryDimension(origin) + getSecondaryDimension(size) );

         }

         int getLeastPrimaryDimension() const    { return getPrimaryDimension(upperLeft_);  }  // in the box
         int getGreatestPrimaryDimension() const { return getPrimaryDimension(lowerRight_); }  // outside the box

         int getLeastSecondaryDimension() const    { return getSecondaryDimension(upperLeft_);  }  // in the box
         int getGreatestSecondaryDimension() const { return getSecondaryDimension(lowerRight_); }  // outside the box

         bool overlaps(ScreenLocation const &point) const
            /// @return true if the point is in the box -- including sitting along the left and top edges
            /// but not including sitting along the right and bottom
         {
            int primary   = getPrimaryDimension(point);
            int secondary = getSecondaryDimension(point);

            return     (primary   >= getLeastPrimaryDimension()   && primary   < getGreatestPrimaryDimension)
                  &&   (secondary >= getLeastSecondaryDimension() && secondary < getGreatestSecondaryDimension)
                  ;
         }

         template<class U>
         static inline bool between(U const &a, U const &l, U const &r)
             ///  @return true if a test value, @b a, is greater than or equal to @b l, and strictly less
             ///  than @b r.
         {
            return a >= l && a < r;      // intentionally closed on the left and open on the right
         }

         bool overlaps(Box const &rhs) const 
             ///  return true if *this overlaps another box.
         {
              int myLeft      = getLeastPrimaryDimension();
              int myRight     = getGreatestPrimaryDimension();    // not "in" me
              int myTop       = getLeastSecondaryDimension();
              int myBottom    = getGreatestSecondaryDimension();  // not "in" me

              int rhsLeft     = rhs.getLeastPrimaryDimension();
              int rhsRight    = rhs.getGreatestPrimaryDimension(); // not "in" rhs
              int rhsTop      = rhs.getLeastSecondaryDimension();  
              int rhsBottom   = rhs.getGreatestSecondaryDimension(); // not "in" rhs

              return (  (    (    between(myLeft,      rhsLeft, rhsRight)
                               || between(myRight-1,   rhsLeft, rhsRight)
                             )
                         &&  (
                                  between(myTop,       rhsTop, rhsBottom)
                               || between(myBottom-1,  rhsTop, rhsBottom)
                             )
                        )
                     || (    (    between(rhsLeft,     myLeft, myRight)
                               || between(rhsRight-1,  myLeft, myRight)
                             )
                         &&  (    between(rhsTop,      myTop, myBottom)
                               || between(rhsBottom-1, myTop, myBottom)
                             )
                        )
                     )
                     ;

         }


         void includePrimaryDimensionLocation(int location) 
             /// Expand the size of the box such the specified location is
             /// now considered to be actually in the box
         {
              if(location < getPrimaryDimension(upperLeft_))
              {
                 setPrimaryDimension(upperLeft_, location);
              }
              else
              if(location >= getPrimaryDimension(lowerRight_))
              {
                setPrimaryDimension(lowerRight_, location+1);   // the plus 1 is needed because
                                                                // the lower right primary dimension
                                                                // is not actually in the box
              }


         }

         void includeSecondaryDimensionLocation(int location) 
             /// Expand the size of the box such the specified location is
             /// now considered to be actually in the box
         {
              if(location < getSecondaryDimension(upperLeft_))
              {
                 setSecondaryDimension(upperLeft_, location);
              }
              else
              if(location >= getSecondaryDimension(lowerRight_))
              {
                setSecondaryDimension(lowerRight_, location+1);   // the plus 1 is needed because
                                                                  // the lower right primary dimension
                                                                  // is not actually in the box
              }


         }


     };

     template<class ViewPortContainer>
     static ViewPort* hasViewport(ViewPortContainer &viewports, int originCol, int endCol)
        /// @return true if there is at least one window with the specified begin end std::pair of columns
     {
     
        CXXTLS_FOREACH(ViewPort *cur, viewports)
        {
            if(   (getPrimaryDimension(getOrigin(cur)) == originCol)
               && (getPrimaryDimension(getOrigin(cur)) + getPrimaryDimension(getSize(cur)) >= endCol)
              )
            {
               return cur;
            }
     
        }
     
        return 0;  // not found
     }
     
     
     template<class ViewPortContainer>
     static bool widenViewport(ViewPortContainer &viewports, 
                               ViewPort             *beneficiary,
                               Manager              *manager
                              )
         /// Widen (or make taller) a viewport and anything else that needs it, then return true if successful.
         ///
         /// Depending on the template parameter, Config, the @b beneficiary will be widened or made taller.
         ///
         /// @param viewports   -- a list/vector/deque of "sub-windows" of a "screen"
         /// @param beneficiary -- one of the sub-windows on the screen.  It will be made wider or taller if possible.
         /// @param manager     -- a debug parameter which lets you get the name of the sub-window (ie ViewPort)
         ///                       for debug purposes.  You can have the Manager::viewerFullname() return "" if you
         ///                       don't want to implement it otherwise.
     {
     
         std::set< int > sorter; // used to sort interesting column values: origin and end columns for windows
         
         ScreenLocation const &beneficiaryOrigin = beneficiary->origin();
         ScreenLocation const &beneficiarySize   = beneficiary->size();
         int                   beneficiaryEnd    = getPrimaryDimension(beneficiaryOrigin) + getPrimaryDimension(beneficiarySize);

         // STEP 1:
         // Put all interesting rows (or columns) in the sorter.  Interesting rows include the top and the bottom
         // of windows (or left and rigth, depending on which is the primary dimension).

         CXXTLS_FOREACH(ViewPort *cur, viewports)
         {
             ScreenLocation const origin = getOrigin(cur);
             ScreenLocation const size   = getSize(cur);
     
             int endCol = getPrimaryDimension(origin) + getPrimaryDimension(size);
     
             sorter.insert( getPrimaryDimension(origin) );
             sorter.insert( endCol );
         }
     
         // we now have one copy of all interesting column numbers in our set, in ascending order.  The set
         // includes the top and the bottom both or the left and right both, depending on which is the 
         // primary dimension.
     
         typedef std::pair<int,int> InterestingLocation;   // A pair of interesting locations:  the first
                                                           // member is the current position of the interesting
                                                           // location stored in the sorter vector.  The second
                                                           // members is the place to which that location has
                                                           // moved.
     
         std::vector< InterestingLocation > interestingLocations;
     
         CXXTLS_FOREACH(int cur, sorter)
         {
            interestingLocations.push_back( std::pair<int,int>(cur,cur) );
         }
     
         // within std::pair InterestingLocation, the first member refers to the old interesting location
         // and the second member refers to the new (ie moved) value.  The first member serves as a
         // search key, and the new value serves the location the window has widened or moved to.
     
         // STEP 2:
         // Search for a donor column to be moved 1 column (or row to be moved 1 row) to allow the
         // beneficiary to be widened.
         //
         // Note that the vector interesting locations contains the beginning and ending points of each
         // viewport -- but only once.  That is, if a windows overlap, they will not have multiple copies
         // of the same numbers in the interestingLocations vector.  For example, consider the following
         // two viewports ( 1 and 2):
         //
         //   a        b     c
         //    +-------+----+
         //    |   1   | 2  |
         //    +-------+----+
         //
         // If columns are the interesting dimension, then interestingLocations would have the
         // following values:
         //
         //     (a,a),(b,b),(c,c)
         //
     
         if(interestingLocations.size() < 3)
         {
            return false; // fail because you have to have at least 2 locations to define the beginning
                          // and ending of a window (origin and origin+size).  But you also have to have
                          // a third location because you can't wident a single window.  If you don't have
                          // at least two horizontal windows, you can't widen 1 with respect to the other.
         }
     
         size_t donor = -1;
         int    donorWidth=0;

         ViewPort *donator=0;
     
         for(size_t i = 0; i < interestingLocations.size() -1; ++i)
         {
             int cur  = interestingLocations[i].first;
             int next = interestingLocations[i+1].first;  // loop iteration stops early so this never an error
     
             // the candidate donor window lies cur->next
     
             if(    (cur < getPrimaryDimension(beneficiaryOrigin) && next <= getPrimaryDimension(beneficiaryOrigin))
                ||  (cur >= beneficiaryEnd       && next >= beneficiaryEnd       )
               )
             {
                // this guy is a possible donor

                ViewPort *candidateDonator;
     
                if(   (next-cur >= Config::minSize)
                   && (candidateDonator = hasViewport(viewports, cur, next))
                   && (donorWidth < (next-cur))
                  )
                {
                   donorWidth = next-cur;
                   donor = i;
                   donator = candidateDonator;
                }
             }
         }     
     
         if(donor == size_t(-1))
         {
            return false;  // no potential donor correctly sized or positioned 
         }

         if(!donator)
         {
            return false; // this should never happen, but who knows
         }



         // STEP3:  
         // Adjust all the interesting locations (member second) such that after the movement
         // the beneficiary viewport is now wider, the donor is smaller, and all viewports in
         // between are moved on the screen to account for the change.  No gaps are allowed
         // to be created.
         //
         // Note that this affects only the contents of variable interestingLocations[*].second.
         // the application of the change to actual viewports occurs earlier.


         // adjust column width for the donor and beneficiary and slide columns around as neded
         {
             size_t beneficiaryIndex;
             int    beneficiaryOriginCol = getPrimaryDimension(beneficiaryOrigin);
     
             for(beneficiaryIndex = 0; beneficiaryIndex < interestingLocations.size() -1; ++beneficiaryIndex)
             {
                if(beneficiaryOriginCol == interestingLocations[beneficiaryIndex].first)
                   break;
             }
     
             if(beneficiaryIndex == interestingLocations.size()-1)
             {
                return false;
             }
     
             if(donor < beneficiaryIndex)
             {
                // donor lies before beneficiary
     
                for(size_t i = donor+1; i <= beneficiaryIndex; ++i)
                   interestingLocations[i].second -= 1; // minus size of donation  
     
                // space has just been removed from the donor and added to the benficiary
                // because the end of the beneficiary column has not changed but the start has
     
     
             }
             else
             {
                // donor lies after beneficiary
     
                for(size_t i = beneficiaryIndex +1; i <= donor; ++i)
                   interestingLocations[i].second += 1; // plus the size of hte donation
     
                // note that the donor has just lost space space to the beneficiary
     
             }
     
         }

         // STEP 4: 
         // Determine which windows should actually be moved by creating a rectangle that exactly
         // includes the region of the screen that should be affected.  This rectangle will be used
         // later to filter out undesirable effects on windows that should not be moved.
         //
         // Explanation:  In order to prevent gaps from appearing in the tiles, or from having
         // multi-part windows, the above algorithm moves more windows around than are actually
         // necessary.  For example:
         //
         //   +-----------+------+--------------+
         //   |           |      |              |
         //   |    a      |  b   |      c       |
         //   |           |      |              |
         //   +-----------+      +--------------+
         //   |    e      |      |     d        |
         //   +-----------+------+--------------+
         //
         // Suppose we are growing window d vertically.  The algorithm above, will move the
         // line above d upward, but will also move the line above e upward because they
         // share the same row value.  We do not really want e to grow just because d is
         // growing.  So the following algorithm creates a rectangle that included d and c
         // but not a, b, or e.
         //
         // It includes all rows that overlap d and whose columns overlap d.  Well, first
         // d and then d + c.  Had b been involved, the affectedRegion would include
         // d + c + b, but it is not.  


         Box affectedRegion(beneficiary);    // the totaliy of the screen region to be
                                             // affected by this widenViewport() call --
                                             // it starts with the beneficiary obviously, and
                                             // includes anthing that it touches which 
                                             // shares secondary dimensions of interest.

         bool changedAffectedRegion;         // flag used to tell us when to stop iterating

         std::set<ViewPort *> affectees;     // set of viewports affected by the change.  This
                                             // is used to tell us that a viewport has already
                                             // been included and that we don't need to reconsider it.

         do
         {
            changedAffectedRegion=false;

            CXXTLS_FOREACH(ViewPort *vp, viewports)
            {
                std::string fullname = manager->viewerFullname(vp);  // for debug purposes only

                if(affectees.find(vp) != affectees.end())
                   continue; 

                Box cur(vp);

                // check to see if the current viewport begins or ends on the primary dimension of the
                // currently active region.  If it does, and if it overlaps the activate region in the
                // secondary dimension, then include it in the active region.


                if(    (cur.getLeastPrimaryDimension()    == affectedRegion.getLeastPrimaryDimension())
                   ||  (cur.getGreatestPrimaryDimension() == affectedRegion.getGreatestPrimaryDimension())
                   ||  (cur.getLeastPrimaryDimension()    == affectedRegion.getGreatestPrimaryDimension())
                   ||  (cur.getGreatestPrimaryDimension() == affectedRegion.getLeastPrimaryDimension())
                  )
                {
                    // current viewport begins or ends at the affected region in the primary dimension.

                    int affLeastSecondary    = affectedRegion.getLeastSecondaryDimension();
                    int affGreatestSecondary = affectedRegion.getGreatestSecondaryDimension();
                    int myLeastSecondary     = cur.getLeastSecondaryDimension();
                    int myGreatestSecondary  = cur.getGreatestSecondaryDimension();

                    if(
                             Box::between(myLeastSecondary,       affLeastSecondary,  affGreatestSecondary)
                          || Box::between(myGreatestSecondary-1,  affLeastSecondary,  affGreatestSecondary)
                          || Box::between(affLeastSecondary,      myLeastSecondary,   myGreatestSecondary)
                          || Box::between(affGreatestSecondary-1, myLeastSecondary,   myGreatestSecondary)
                      )
                    {
                        // Affected region overlaps the current viewport in the secondary dimension -- not including
                        // the end of range

                        affectees.insert(vp);

                        changedAffectedRegion=true;

                        affectedRegion.includePrimaryDimensionLocation(cur.getLeastPrimaryDimension());
                        affectedRegion.includePrimaryDimensionLocation(cur.getGreatestPrimaryDimension()-1);

                        affectedRegion.includeSecondaryDimensionLocation(myLeastSecondary);
                        affectedRegion.includeSecondaryDimensionLocation(myGreatestSecondary-1);

                    }

                }

            }

         }
         while(changedAffectedRegion);
     
         // STEP 5:
         // Apply needed deltas to all windows that are inclosed in the affectedRegion.

         // Apply needed deltas to all windows.  Note that the "first" member of an InterestingLocation is
         // unchanged at this point and only some "second" members.
         //
         // Note that this application changes some windows that aren't technically necessary but eliminating
         // them would be very difficult.  Let's see if it is truly necessary.
     
         CXXTLS_FOREACH(ViewPort *vp, viewports)
         {
            ScreenLocation oldOrigin = vp->origin();
            ScreenLocation oldSize   = vp->size();
            int                   oldEnd    = getPrimaryDimension(oldOrigin) + getPrimaryDimension(oldSize);
     
            std::string fullname = manager->viewerFullname(vp);

            Box vpRegion(oldOrigin, oldSize);

            if(!vpRegion.overlaps(affectedRegion))
                continue;

            size_t scan;

            // if this window is in the affected region and if it has an edge that has been
            // moved above, then apply the change tot he window.

     
            for(scan = 0; scan < interestingLocations.size() -1; ++scan)
            {
                int cur = interestingLocations[scan].first;
     
                size_t nextSearch = scan+1;
     
                int next= interestingLocations[nextSearch].first;
     
                while(   (next < oldEnd)
                      && (nextSearch < interestingLocations.size() -1)
                     )
                {
                   ++nextSearch;
                   next = interestingLocations[nextSearch].first;
                }
     
     
                if(   (cur  == getPrimaryDimension(oldOrigin))
                   && (next == oldEnd)
                  )
                {
                   // this viewport shares start and end column with the current std::pair of interesting locations
     
                   ScreenLocation newOrigin = oldOrigin;
                   ScreenLocation newSize   = oldSize;
     
                   int newCur = interestingLocations[scan].second;
                   int newNext= interestingLocations[nextSearch].second;
     
                   setPrimaryDimension(newOrigin, newCur);
                   setPrimaryDimension(newSize, newNext - newCur);
     
                   move(vp, newOrigin, newSize);
                }
     
            }
            
         }
     
     
         return true; // success
     
     }

};  // struct ViewportAdjust

namespace WindowAdjustments
{
    struct Span
        ///  A 1D span of space.
    {
        int origin_;
        int width_;

        Span(int origin, int width)
        :  origin_(origin)
        ,  width_ (width)
        {
        }
    };

    struct ResizeInfo
        /// 2 1D spans of space.  These represnet the old span before the resize and the new span.
    {
       Span currentSpan_;
       Span originalSpan_;
       void *id_;

       ResizeInfo(void *id, int origin, int width)
       :  currentSpan_ (origin, width)
       ,  originalSpan_(origin, width)
       ,  id_(id)
       {
       }

       ResizeInfo(void *id, int origin, int width, int oldOrigin, int oldWidth)
       :  currentSpan_ (origin, width)
       ,  originalSpan_(oldOrigin, oldWidth)
       ,  id_(id)
       {
       }


    };

    typedef std::multimap<int, ResizeInfo> ResizeMap;  ///<  A sorted collection of resizable spans.  The
                                                       ///   first template parameter is the origin of the
                                                       ///   original span, before resize.


};

void resizeAllViewports(WindowAdjustments::ResizeMap & currentSizes,
                        int newTotalWidth
                       );
    ///<  Given a sorted collection of window viewport spans, handle an overall change
    ///  to the size.




} // namespace cxxtls
#endif
