#ifndef classTraits_header_included_p
#define classTraits_header_included_p
//
// Copyright 2008, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.  Specifically the
// "guts" of the is_polymorphic<T> mechanism comes from boost -- although
// I do feel like an idiot for not having thought of it myself.  Ok, after
// careful consideration of exactly how the algorithm works, including the
// _selector mechanism, which is a work of genius (appearing in the STL)
// I no longer feel like an idiot.  Whether I am or not, is an exercise left
// to the observer...

/// @file cxxtls/classTraits.h
/// This file makes definitions that are useful to template writers and only
/// marginally useful for any other purpose.  If you are not writing templates
/// you most likely already know the things about your data types
/// that the class queries provided below are meant to tell you.
///
/// @sa classTraits_documentation for comments about the use of the declarations
/// in this file.  Look at file test_classTraits.c for example uses
/// of these types and expected values given various input data types.


#include <memory>  // to get size_t

namespace cxxtls
{

// And now to the class type traits helpers...

//
// the true and false types exist only to aid in template function selection.
//
//  Basically you can include a nested typedef like this in your own class:
//
//   typedef true_value_type  flagType;
//
//  Then when templates are instantiated using your class, they can look at the flag type
//  and make decisions as to whether flagType is typedef'd to true_value_type or false_value_type.
//

struct true_value_type 
  ///
  /// a type that means yes.  
  /// true_value_type::value == 1.  
  /// @see @ref  classTraits_documentation
  ///
{
  typedef true_value_type type;

  enum constants { value = 1 };              // compile time constant usable in EvalTypeIf
                                             // or regular code.
};

struct false_value_type 
  ///
  /// a type that means no.  
  /// false_value_type::value == 1.  
  /// @see @ref  classTraits_documentation
  ///
{
  typedef false_value_type type;

  enum constants { value = 0 };              // compile time constant usable in EvalTypeIf
                                             // or regular code.
};

struct not_a_type{};  // a type that represents "not a type"

template<bool, class ZeroClass, class OneClass> 
struct  EvalTypeIf
  ///
  ///  A mechanism for selecting between either of two types based on a 
  ///  compile time constant expression.
  ///
  ///  The EvalTypeIf&lt;a,b,c&gt; template allows you to select one of two type definitions (b or c)
  ///  based on boolean value (a).  To use it, you create some compile time expression
  ///  that evaluates to true or false, and then use that expression to select either
  ///  of the two type definitions:
  ///
  ///  @verbatim
  ///
  ///  For example:
  ///
  ///     typedef typename EvalTypeIf<0, firstPossibleType, secondPossibleType>::type MyType;
  ///
  ///  selects the firstPossibleType and
  ///
  ///     MyType::type is firstPossibleType.
  ///
  ///  If the first template parameter is 1, then 
  ///
  ///     MyType::type is secondPossibleType
  ///
  ///  @endverbatim
  ///
  ///  In practice, specializations define the possiblities, and a compile time constant
  ///  expression selection among the choices.  For example:
  ///
  ///  @code
  ///    typedef EvalTypeIf< sizeof(int) > sizeof(char), char, int>::type ValueType;
  ///    //
  ///    // here, ValueType will be 'int' because sizeof(int) > sizeof(char)
  ///    //
  ///  @endcode
  ///
  ///  More often, instead of using explict constants (0 or 1) or even explict boolean
  ///  expressions for the selector (argument 1), you will use one of the template classes
  ///  that define compile time constants based on their arguments.  For example, argument
  ///  1 to EvalTypeIf will most like be one of:
  ///
  ///  @code
  ///     isSameType<A,B>::value
  ///     isConstType<A>::value
  ///     isConvertibleType<From,To>::value
  ///     isClassType<A>::value
  ///     isPointerType<A>::value
  ///     isPodType<A>::value
  ///     isPolymorphicType<A>::value
  ///     isSignedType<A>::value
  ///     isUnsignedType<A>::value     // does not include floats or enums
  ///     isArithmeticType<A>::value
  ///     isEnumType<A>::value
  ///     isFloatType<A>::value
  ///     isIntegralType<A>::value
  ///     isEnumType<A>::value
  ///     isTemplate<A>::value
  ///     basicClassType<A>::value  -- defines a unique integer identifier for basic kind type
  ///                                  in the C++ language.  See enumeration @ref cxxtls::basicClassTypes, below.
  /// 
  ///  @endcode 
  ///
  ///  @see @ref isPolymorphicType_selector template below for a realistic example use of EvalTypeIf
  ///
  ///  @see classTraits_documentation for more info on how this works.
  ///
  ///  @note  If you you get compile errors because one or the other of the alternative
  ///         types being selected won't compile, you can use the "_selector" strategy
  ///         below instead of the EvalTypeIf strategy.  @see  isPolymorphicType_selector.
{
  // the default implementation is unusable.
};

template<class ZeroClass, class OneClass> 
struct  EvalTypeIf<0, ZeroClass, OneClass>
{
  typedef ZeroClass type;  // the false case:
};

template<class ZeroClass, class OneClass> 
struct  EvalTypeIf<1, ZeroClass, OneClass>
{
  typedef OneClass type;  // the true case
};

template<class T, class U>
struct isSameType
  ///
  ///  Member value tells you whether two types are the same.
  ///
  ///  You use it like this:
  ///
  ///  @verbatim
  ///
  ///     typename EvalTypeIf< IsSameType<A,B>::value, OneType, OtherType>::type variable;  
  ///        //
  ///        // Variable is of type OneType, or OtherType depending on whether type A and B
  ///        // are the same type.
  ///
  ///  @endverbatim
  ///
  ///  Or maybe like this: 
  ///
  ///  @verbatim
  ///
  ///     if( IsSameType<A,B>::value )
  ///     {
  ///       // A and B are the same type
  ///       ...
  ///     }
  ///
  ///  @endverbatim
  ///
  /// @see @ref  classTraits_documentation
{
  // neither of these functions is ever called or instantiated

  template<class V> static char is_same_type_func(V&,V&);  // this one only takes a pair of ptrs
                                                           // which are equal
  static int is_same_type_func(...);                       // this one takes any number of any type

  static T t;
  static U u;

  // if the two types of template parms are the same, the first func will be called  in the
  // following constant expression -- except that it won't ACTUALLY be called

  enum { value =    (sizeof( is_same_type_func(t,u) ) == sizeof(char))
                 && (sizeof( is_same_type_func(u,t) ) == sizeof(char))
       };

};


template<class T>
struct isConstType
  ///
  ///  See isConstType<T>::value for a const expression telling you whether or not
  ///  T is a "const" type or not.
  ///
  ///  The following templates, removeConst<T>, removeVolatile<T>, and removeCV<T> let you
  ///  define a typedef based on some specified type BUT not having the const or volatile
  ///  attributes.  Use it like this:
  ///
  ///  @verbatim
  ///
  ///    template<class T>
  ///    void function(T t)
  ///    {
  ///       typedef typename removeCV<T>::type PlainOldT;
  ///
  ///       // use PlainOldT instead of T to eliminate annoying and unnecessary
  ///       // template specializations of T which have const or volatile attributes.
  ///    }
  ///
  ///  @endverbatim
  ///
  /// @see @ref  classTraits_documentation
{
  typedef T type;

  typedef false_value_type valueType; 

  enum constants { value = 0 };

};

template<class T>
struct isConstType<const T>
{
  typedef T type; // for const, remove the const

  enum constants { value = 1 };

  typedef true_value_type valueType; 

};

template<class T, size_t N>
struct isConstType<const T[N]>
{
  typedef T type; // for const, remove the const

  enum constants { value = 1 };

  typedef true_value_type valueType; 

};


template<class T>
struct removeConst
  ///
  ///  typename removeConst<T>::type is a type that is the non-const version of T.  For normal
  ///  classes, removeConst<T>::type is the same as T, but for T which is const, type is the
  ///  non-const form.
  ///
  ///  The following templates, removeConst<T>, removeVolatile<T>, and removeCV<T> let you
  ///  define a typedef based on some specified type BUT not having the const or volatile
  ///  attributes.  Use it like this:
  ///
  ///  @verbatim
  ///
  ///    template<class T>
  ///    void function(T t)
  ///    {
  ///       typedef typename removeCV<T>::type PlainOldT;
  ///
  ///       // use PlainOldT instead of T to eliminate annoying and unnecessary
  ///       // template specializations of T which have const or volatile attributes.
  ///    }
  ///
  ///  @endverbatim
  ///
  /// @see @ref  classTraits_documentation
{
  typedef typename isConstType<T>::type type;
};



template<class T>
struct removeVolatile
  ///
  ///  typename removeVolatile<T>::type is a type that is the non-volatile version of T.  For normal
  ///  classes, removeVolatile<T>::type is the same as T, but for T which is const, type is the
  ///  non-const form.
  ///
  ///  The following templates, removeVolatile<T>, removeConst<T>, and removeCV<T> let you
  ///  define a typedef based on some specified type BUT not having the const or volatile
  ///  attributes.  Use it like this:
  ///
  ///  @verbatim
  ///
  ///    template<class T>
  ///    void function(T t)
  ///    {
  ///       typedef typename removeCV<T>::type PlainOldT;
  ///
  ///       // use PlainOldT instead of T to eliminate annoying and unnecessary
  ///       // template specializations of T which have const or volatile attributes.
  ///    }
  ///
  ///  @endverbatim
  ///
  /// @see @ref  classTraits_documentation
{
  typedef T type; // for non-volatle, just use the class 
};


template<class T>
struct removeVolatile<volatile T>
{
  typedef T type; // for volatile, remove it
};

template<class T>
struct removeCV
  ///
  /// typename removeCV<T>::type is the same as type T but is neither const nor is it volatile.
  ///
  /// @see @ref  classTraits_documentation
{
  typedef typename removeConst< typename removeVolatile<T>::type >::type type;
};

//
//  removeReference<T>  has a member, type, which is T without the &, if any
//

template<class U> struct removeReference_impl { typedef U type; };
template<class V> struct removeReference_impl <V&> { typedef V type; };

template<class T>
struct removeReference
    ///
    ///  typename removeReference<T>::type is the same as type T, but is not a reference. 
    ///
    /// @see @ref  classTraits_documentation
{

  typedef typename removeReference_impl<T>::type type; 

};



//
// template isConvertibleType<FromType, ToType> tells you whether or not
// FromType is assignment compatible with ToType, that is, can you construct a
// ToType object from a FromType object.
//


template<class From, class To>
struct isConvertibleType
    ///
    ///  isConvertible<From,To>::value is a compile time constant which is true
    ///  if type From can be converted into type To without special syntax.
    ///
    ///  For example an character can be converted into an integer, but a
    ///  a pointer cannot be converted into an integer without a cast.
    ///
    /// @see @ref  classTraits_documentation
{
   // these functions and variables are never instantiated

   static char f(...);  // this function will take anything as a parm
   static int  f(To); // this function will only take a ToType object as a parm

   // note that the f(...) and f(To&) forms return different data types.  

   static From fromp;  
   static To   top;

   // The following code defines a constant named "value" whose value is defined
   // by the size of the data that would be returned by one of the two 'f' overloads
   // above.  No actual call is made, but the compiler can tell which of the two
   // overloaded forms WOULD be called with the given arguments and thus it can tell
   // the return type -- and thus you can take the size thereof and compare the size of
   // the returned value of the two different function calls.
   //
   // 
   // If the from and two classes are assignment compatible then the f(To&) form will
   // be called.  If they are not, then different functions will be called -- and thus
   // the size of the return value will be different.  Of course, they are not REALLY called
   // only evaluated for call...

   enum constants
   {
     value = sizeof( f(fromp) ) ==  sizeof( f(top) )
   };


};

template<class To>
struct isConvertibleType<void, To>
   ///  A wrapper for a value that indicates that type "To" is a void.
{
  enum { value = false };
};
template<class To>
struct isConvertibleType<const void, To>
   ///  A wrapper for a value that indicates that type "To" is a const void.
{
  enum { value = false };
};

template<>
struct isConvertibleType<float, int>
{
  enum { value = true };
};
template<>
struct isConvertibleType<const float, int>
{
  enum { value = true };
};
template<>
struct isConvertibleType<double, int>
{
  enum { value = true };
};
template<>
struct isConvertibleType<const double, int>
{
  enum { value = true };
};
template<>
struct isConvertibleType<long double, int>
{
  enum { value = true };
};
template<>
struct isConvertibleType<const long double, int>
{
  enum { value = true };
};



template<class Type>
struct isClassType
    ///
    ///  isClassType<T>::value is a compile time constant that is true only if T is a class, or struct.
    ///
    /// @see @ref  classTraits_documentation
{
  
  template <typename ClassType>
  static short fun(void (ClassType::*)());
  
  template <typename NonClassType>   // this has to be a template
  static char fun(...);

  enum { value = sizeof(fun<Type>(0)) == sizeof(short) };
};


template<class T>
struct isFunctionType
  ///
  ///  isFunctionType<T>::value is a compile time constant that is only true only if T can be executed like a function of 0 to 9 parameters.
  ///
  /// @see @ref  classTraits_documentation
{
  // works on functions up to 9 parameters -- feel free to add more

  template<class RV> static char f(RV (*func0)());
  template<class RV, class A> static char f(RV (*func1)(A));
  template<class RV, class A, class B> static char f(RV (*func2)(A, B));
  template<class RV, class A, class B, class C> static char f(RV (*func3)(A, B, C));
  template<class RV, class A, class B, class C, class D> static char f(RV (*func4)(A, B, C, D));
  template<class RV, class A, class B, class C, class D, class E> static char f(RV (*func5)(A, B, C, D, E));
  template<class RV, class A, class B, class C, class D, class E, class F> static char f(RV (*func6)(A, B, C, D, E, F));
  template<class RV, class A, class B, class C, class D, class E, class F, class G> static char f(RV (*func7)(A, B, C, D, E, F, G));
  template<class RV, class A, class B, class C, class D, class E, class F, class G, class H> static char f(RV (*func8)(A, B, C, D, E, F, G, H));
  template<class RV, class A, class B, class C, class D, class E, class F, class G, class H, class I> static char f(RV (*func9)(A, B, C, D, E, F, G, H, I));

  static int f(...);

  static T t;

  enum { value = sizeof(char) == sizeof( f(t) ) };

};

template<>
struct isFunctionType<void>
{
  enum { value = false };
};

template<>
struct isFunctionType<const void>
{
  enum { value = false };
};



template<class T>
struct isPointerType
    ///
    ///  isPointerType<T>::value is a compile time constant that is true only if T is a pointer
    ///  or a builtin array -- which can be treated like a pointer.
    ///
    /// @see @ref  classTraits_documentation
{
  enum constants { value = 0 };
};

template<class T>
struct isPointerType<T*>
{
  enum constants { value = 1 };
};

template<class T>
struct isPointerType<T const*>
{
  enum constants { value = 1 };
};

template<class T>
struct isPointerType<T* const>
{
  enum constants { value = 1 };
};

template<class T>
struct isPointerType<T const* const>
{
  enum constants { value = 1 };
};

template<class T, size_t N>
struct isPointerType<T [N]>
{
  enum constants { value = 1 };
};

template<class T, size_t N>
struct isPointerType<T const [N]>
{
  enum constants { value = 1 };
};


template<class T>
struct isArrayType
    ///
    /// isArrayType<T>::value is true if T is an array
    ///
{
   enum constants { value = 0 }; 
};


template<class T, size_t N>
struct isArrayType<T [N]>
    ///
    /// isArrayType<T>::value is true if T is an array
    ///
{
  enum constants { value = 1 };
};

template<class T, size_t N>
struct isArrayType<T const [N]>
    ///
    /// isArrayType<T>::value is true if T is an array
    ///
{
  enum constants { value = 1 };
};



//
//  typename pointerReferent<T>::type is a type gives you the base type from a pointer type.
//
//  non-pointers are their own pointee type


template<class T>
struct pointerReferent
    ///
    ///  pointerReferent<T>::type is the type that T points to if T is a pointer.  If T is not
    ///  a pointer, then pointerReference<T>::type is the same as T.
    ///
    /// @see @ref  classTraits_documentation
{
  typedef T type; // non-pointers should not cause compile errors since you can use isPointer
                  // at run time to tell if it really is a pointer or not
};

template<class T>
struct pointerReferent<T*>
   /// Determine the type referred to by a mutable pointer.  See member "type".
{
  typedef T type;
};

template<class T>
struct pointerReferent<T const*>
   /// Determine the type referred to by a const pointer.  See member "type".
{
  typedef T const type;
};

template<class T, size_t N>
   /// Determine the type referred to by an array's name.  See member "type".
struct pointerReferent<T [N]>
{
  typedef T type;
};

template<class T, size_t N>
struct pointerReferent<T const [N]>
   /// Determine the type referred to by a const array's name.  See member "type".
{
  typedef T const type;
};


template<class T>struct ispodtype_impl: public false_value_type {};  
    ///< Generic template that defines a member, value, which is non-zero if
    ///  the template parameter, T, is a POD type object -- that is, if T is
    ///  plain old data, then value is true.


// only the following are pod:

template<class T> struct ispodtype_impl<T*>          : public true_value_type {}; ///< pointers are POD
template<class T> struct ispodtype_impl<T* const>    : public true_value_type {}; ///< const pointers are POD
template<> struct ispodtype_impl<char>               : public true_value_type {}; ///< char is POD
template<> struct ispodtype_impl<short>              : public true_value_type {}; ///< short is POD
template<> struct ispodtype_impl<int>                : public true_value_type {}; ///< int is POD
template<> struct ispodtype_impl<long>               : public true_value_type {}; ///< long is POD
template<> struct ispodtype_impl<long long>          : public true_value_type {}; ///< long long is POD
template<> struct ispodtype_impl<unsigned char>      : public true_value_type {}; ///< unsigned's are all POD
template<> struct ispodtype_impl<unsigned short>     : public true_value_type {}; ///< unsigned's are all POD
template<> struct ispodtype_impl<unsigned int>       : public true_value_type {}; ///< unsigned's are all POD
template<> struct ispodtype_impl<unsigned long long> : public true_value_type {}; ///< unsigned's are all POD
template<> struct ispodtype_impl<float>              : public true_value_type {}; ///< float's are all POD
template<> struct ispodtype_impl<double>             : public true_value_type {}; ///< float's are all POD

// arrays are pod


template<class T>
struct isPodType
  ///
  ///  isPodType<T>::value is a compile time constant that is true if T is a plain old data
  ///  type and not a struct or union.
  ///
  /// @see @ref  classTraits_documentation
{
  enum { value = ispodtype_impl<T>::value || isConvertibleType<T,int>::value || isFunctionType<T>::value };
};

template<class T, size_t N> struct ispodtype_impl<T [N]> : public isPodType<T> {};  ///< arrays are pod if their contents are pod


//
//  The isPolymorphicType<T> template lets you find out if a class has virtual methods.
//  Note that its purpose is to define nested class members which give you information
//  about T -- is it polymorphic or not.
//



template <class plainT>
struct isPolymorphicType_impl
  //
  //  This definition handles non-POD types
  //
{
  // to tell if a data type is polymorphic, you check to see if it has a virtual
  // function table.  The way you tell that is to derive new classes from it.  You
  // derive a new class with a new virtual method and a different class without any
  // new virtual methods.  If if sizeof the new types is equal then you know that the
  // base class has virtual methods to start with, but if the sizes are not the same
  // then the it was not originally polymorphic.


   struct d1 : public plainT
   {
      d1();
      ~d1()throw();
      char padding[256];
   };

   struct d2 : public plainT
   {
      d2();
      virtual ~d2() throw();
      char padding[256];
   };

   enum constants { value = sizeof(d1) == sizeof(d2) };

   typedef typename EvalTypeIf<value, false_value_type, true_value_type>::type type;

   operator bool() const { return value; }

};

/// isPolymorphicType_selector exists to work around a compile error problem:
///
/// If you use a type, it must be syntactically correct -- unless it is a template which
/// does not always have to.
/// 
/// The error occurred when we tried to use the isPolymorphicType_impl<enumType>.  Since
/// there are an infinite number of enum types, we can't specially them all.  The work around
/// is to check for a class object before attempting to instantiating isPolymorphicType_impl<T>.
/// In this way, if the data type isn't a class, we never instantiate the offending type.
///
/// To pull this off, a choice of isPolymorphicType_selector<T> is made.  One specialization,
/// for classes, has a nested sub-type template defines a sub-sub-type which invokes 
/// isPolymorphicType_impl<T>.  The other specialization of isPolymorphicType_selector
/// does not invoke isPolymorphicType_impl<T>, but just stops with a false indication.
///
/// The rebind<T> syntax is just a compilicated way of not compiling isPolymorphicType_impl<T> --
/// unless we absolutely need it.

template <bool is_class>
struct isPolymorphicType_selector
{
  // non-class case

   template <class T>
   struct rebind
   {
      typedef false_value_type type;
   };
};

template<>
struct isPolymorphicType_selector<true>
{
  // isClassType case

   template <class T>
   struct rebind
   {
      typedef isPolymorphicType_impl<T> type;
   };
};

template <class T>
struct isPolymorphicType
    ///
    ///  isPolymorphicType<T>::value is a compile time constant that is true
    ///  only if T is a class with virtual methods.
    ///
    ///  The implementation of this class uses a very sophisticated method to
    ///  to determine if T is a pointer, arithmetic, enumeration, etc.  The
    ///  idea is stolen from boost, which is stolen from the iostreams code.
    ///
    /// @see @ref  classTraits_documentation
{
   // Normally, we would just use isPolymorphicType_impl<T>::value, but
   // the isPolymorphicType_impl<T> class would produce a compile error if
   // T is an enumeration, pointer, or arithmetic data type name.
   //
   // The following syntax to select a guaranteed false data type for non-class 
   // values of T, but for true classes, select isPolymorphicType_impl<T> -- 
   // which will compile successfully.

   typedef isPolymorphicType_selector< isClassType<T>::value> selector;

   // since isPolymorphicType_selector<false> has a rebind template that defines
   // a guaranteed false_value_type, we are assured that selector will compile
   // no matter what data type is used for T -- and for class types, which gives
   // defines a selector with a rebind method that refers to isPolymorphicType_impl,
   // we get proper operation for that case too.
   // 

   // the following code invokes a template on the selector type.
   // I'm not sure why you need the template keyword, but you do.


   typedef typename selector::template rebind<T> binder;
   typedef typename binder::type imp_type;
   enum { value = imp_type::value };
};




//
//  this template tells you if a class represents an signed number
//

template<class T>
struct isSignedType
    ///
    ///  isSignedType<T>::value is a compile time constant that is true only if 
    ///  T is a signed number.
    ///
    /// @see @ref  classTraits_documentation
{
  typedef typename removeCV<T>::type mutableType;

  static char f(char &);
  static char f(int &);
  static char f(short &);
  static char f(long &);
  static char f(long long &);
  static char f(float &);
  static char f(double &);
  static char f(long double &);

  static int  f(...);

  static mutableType t;

  enum { value = (sizeof(f(t)) == sizeof(char) ) };


};

template<>
struct isSignedType<void>
{
  enum { value = 0 };
};
template<>
struct isSignedType<const void>
{
  enum { value = 0 };
};

template<class T>
struct isUnsignedType
    ///
    ///  isUnsignedType<T>::value is a compile time constant which is true only if
    ///  T is an unsigned numeric type.
    ///
    /// @see @ref  classTraits_documentation
    ///
    /// Note that enumeration variables and floating point numbers are not "unsigned".
    /// enums can be negative!
{
  typedef typename removeCV<T>::type mutableType;
  typedef typename removeReference<mutableType>::type UnrefType;

  static char f(unsigned char &);
  static char f(unsigned int &);
  static char f(unsigned short &);
  static char f(unsigned long &);
  static char f(unsigned long long&);

  static int  f(...);

  static UnrefType t;

  enum { value =sizeof(f(t)) == sizeof(char) };

};

template<>
struct isUnsignedType<void>
{
  enum { value = 0 };
};
template<>
struct isUnsignedType<const void>
{
  enum { value = 0 };
};



template<class T>
struct isFloatType
    ///
    ///  isFloatType<T>::value is a compile time constant that is only true if
    ///  if T is a floating point number of some form.
    ///
    /// @see @ref  classTraits_documentation
{
  typedef typename removeCV<T>::type mutableType;
  typedef typename removeReference<T>::type unrefType;

  static char f(float &);
  static char f(double &);
  static char f(long double &);
  static int  f(...);

  static unrefType *t;

  enum { value = sizeof(f(*t)) == sizeof(char) };
};

template<>
struct isFloatType<void>
    ///
    ///  A definition of isFloatType for void type.  Void is not a flaot type.
    ///
{
  enum { value = 0 };
};

template<>
struct isFloatType<const void>
    ///
    ///  A definition of isFloatType for const void type.  Void is not a flaot type.
    ///
{
  enum { value = 0 };
};






template<class T>
struct isEnumType
    ///
    ///  isEnumType<T>::value is a compile time constant that is only true if T is 
    ///  an enumeration type -- and not simple arithmetic value, pointer, class, function
    ///  pointer, etc.
    ///
    /// @see @ref  classTraits_documentation
{
  typedef typename removeCV<T>::type mutableType;
  typedef typename removeReference<mutableType>::type UnrefType;

  static char f(unsigned char &);
  static char f(unsigned int &);
  static char f(unsigned short &);
  static char f(unsigned long &);
  static char f(unsigned long long&);
  static char f(char &);
  static char f(int &);
  static char f(short &);
  static char f(long &);
  static char f(long long&);
  static char f(float &);
  static char f(double &);
  static char f(long double &);


  static int  f(...);

  static UnrefType t;

  enum { value =    !isClassType<UnrefType>::value 
                 && isConvertibleType<UnrefType, int>::value 
                 && sizeof( f(t) ) != sizeof(char)
       };
};

template<>
struct isEnumType<void>
{
  enum { value = 0 };
};
template<>
struct isEnumType<const void>
{
  enum { value = 0 };
};



template<class T>
struct isIntegralType
    ///
    /// isIntegralType<T>::value is a compile time constant that is only true if
    /// T is an int, short, char, long, long long, or unsigned versions thereof.
    ///
    /// @see @ref  classTraits_documentation
{
  typedef typename removeCV<T>::type MutableType;
  typedef typename removeReference<MutableType>::type U;


  enum { value = (   isSignedType<U>::value 
                  || isUnsignedType<U>::value 
                 )
                 && !isEnumType<U>::value
                 && !isFloatType<U>::value
                 && !isArrayType<T>::value
       } ;
};



template<class T>
struct isArithmeticType
    ///
    ///  isArithmeticType<T>::value is a compile time constant that is only true if
    ///  T can be used in normal arithmetic operations.
    ///
    /// @see @ref  classTraits_documentation
{
  typedef typename removeCV<T>::type MutableType;
  typedef typename removeReference<MutableType>::type U;


  enum { value = isIntegralType<U>::value || isFloatType<U>::value };
};

template<>
struct isArithmeticType<void>
: public false_value_type
    ///
    ///  isArithmeticType<void> is a variant of isArithmeticType that does not define a member named value
    ///  and thus will will not compile correctly if an arithmetic type is required.
    ///
{
};

template<>
struct isIntegralType<void>
: public false_value_type
    ///
    ///  isIntegralType<void> is a variant of isIntegralType that does not define a member named value
    ///  and thus will will not compile correctly if an integral type is required.
    ///
{
};

template<>
struct isArithmeticType<const void>
: public false_value_type
    ///
    ///  isArithmeticType<void> is a variant of isArithmeticType that does not define a member named value
    ///  and thus will will not compile correctly if an arithmetic type is required.
    ///
{
};

template<>
struct isIntegralType<const void>
: public false_value_type
    ///
    ///  isIntegralType<void> is a variant of isIntegralType that does not define a member named value
    ///  and thus will will not compile correctly if an integral type is required.
    ///
{
};


template<class T>
struct isTemplate // supports up to 9 template parameters (all class)
                  // DOESN'T COMPILE EVERYWHERE!
    ///
    ///  isTemplate<T>::value is a compile time constant that is only true if 
    ///  T is a template of between 1 and 9 template parameters.
    ///
    ///  Note that this class is not overly useful, but it is possible to define it,
    ///  so in the hopes that it might theoretically be useful it is included in the
    ///  library.
    ///
    /// @see @ref  classTraits_documentation
{
  static typename removeCV<T>::type *t;

  static int f(...);  // all types but templates

  template<class A, 
           template<class A1> class Template
          > 
          static char f( Template<A> *);

  template<class A, class B,
           template<class A1, class B1> class Template
          > 
          static char f( Template<A,B> *);

  template<class A, class B, class C,
           template<class A1, class B1, class C1> class Template
          > 
          static char f( Template<A,B,C> *);

  template<class A, class B, class C, class D,
           template<class A1, class B1, class C1, class D1> class Template
          > 
          static char f( Template<A,B,C,D> *);

  template<class A, class B, class C, class D, class E,
           template<class A1, class B1, class C1, class D1, class E1> 
           class Template
          > 
          static char f( Template<A,B,C,D,E> *);

  template<class A, class B, class C, class D, class E, class F,
           template<class A1, class B1, class C1, class D1, class E1, class F1> 
           class Template
          > 
          static char f( Template<A,B,C,D,E,F> *);

  template<class A, class B, class C, class D, class E, class F, class G,
           template<class A1, class B1, class C1, class D1, class E1, class F1, class G1> 
           class Template
          > 
          static char f( Template<A,B,C,D,E,F,G> *);

  template<class A, class B, class C, class D, class E, class F, class G, class H,
           template<class A1, class B1, class C1, class D1, class E1, class F1, class G1, class H1> 
           class Template
          > 
          static char f( Template<A,B,C,D,E,F,G,H> *);

  template<class A, class B, class C, class D, class E, class F, class G, class H, class I,
           template<class A1, class B1, class C1, class D1, class E1, class F1, class G1, class H1, class I1> 
           class Template
          > 
          static char f( Template<A,B,C,D,E,F,G,H,I> *);


  enum { value = sizeof( f(t) ) == 1 };
};


enum basicClassTypes
   /// A list of identifiers for the basic kinds of data types that can
   /// appear in a C++ program -- as defined by the template instantiator.
{
    bctUnknownType  =0,
    bctVoidType     =1,
    bctIntegralType =2,  // int, long, char
    bctEnumType     =3,
    bctPointerType  =4,
    bctArrayType    =5,
    bctFloatType    =6,
    bctFunctionType =7,
    bctClassType    =8,
};

template<class T> struct basicClassType; 

template<> 
struct basicClassType<void> 
   ///
   ///  A type that contains enumeration value, named "value", which
   ///  indicates the basic kind of data that a template parameter T
   ///  refers to.  This particular specialization of the template
   ///  identifies the "void" type which cannot be identified by the
   ///  more complex expression in the main specialization.
   ///
{ 
    enum { value = bctVoidType    }; 
};

template<class T> 
struct basicClassType
    ///
    ///  Defines a member, value, to be one of the @ref basicClassTypes values.
    ///
    ///  This class exists so that you can use the defined value to select implementation
    ///  of a function or class based on the basicType of the template parameter, T.
    ///
    ///  For example, suppose you want to write a template function that can handle 
    ///  any parameter type -- but that you know that different data types must have different
    ///  algorithms.  Here's how you can use the basicClassType template to your advantage:
    ///
    ///@code
    ///
    ///  template<int N, class T> struct alternative_implementations;  
    ///      //
    ///      // alternative_implementations                                                 
    ///      // is a library of implementations of your algorithm based on the
    ///      // type of the template parameter to your function, shown below.
    ///
    ///  template<class T>
    ///  void function(T &t)
    ///    //
    ///    //  Your function just becomes a wrapper around the template trickery
    ///    //  that selects the desired implementation based on the type of the
    ///    //  parameter.
    ///    //
    ///
    ///  {
    ///     alternative_implementations< basicClassType<T>::value, T >::doit(t);
    ///        // Note:  if the above line of code fails to compile it is likely
    ///        // because there is no specialization for this kind of T!
    ///      
    ///  }
    ///
    ///  Here is an example specialization:
    ///
    ///    template<class T> 
    ///    struct alternative_implementations<bctIntegralType, T>
    ///    {
    ///       static void doit(T &t) { ... };
    ///    };
    ///
    ///@endcode
    ///
{
    typedef typename removeCV<T>::type MutableType;
    typedef typename removeReference<MutableType>::type RawType;

    enum {
            value = ( isIntegralType<RawType>::value
                      ? ( isArrayType<T>::value ? bctArrayType : bctIntegralType)  // 2
                      : ( isEnumType<RawType>::value
                          ? bctEnumType  // 3
                          : ( isArrayType<RawType>::value
                              ? bctArrayType // 5
                              : ( isFunctionType<RawType>::value
                                  ? bctFunctionType // 7
                                  : ( isFloatType<RawType>::value
                                     ? bctFloatType // 6
                                     : ( isPointerType<RawType>::value
                                        ? bctPointerType // 4
                                        : ( isClassType<RawType>::value
                                           ? bctClassType
                                           : bctUnknownType  // should not happen. (0)
                                          )
                                       )
                                    )
                                )
                            )
                        )
                    )
         };
};

template<class Base, class Derived>
void derivedFromBase(Base *bptr, Derived *&dref)
    ///  Compute the address of a Derived class object given a
    ///  pointer to a Base class object from which it is derived.
    ///
    ///
    ///  @param bptr  The 'this' pointer from an object which is a
    ///               base class member of a Derived class object.
    ///
    ///  @param dref  A reference to a pointer that will be 
    ///               populated with address of a a sub-structure in the
    ///               base class object.
    ///
    ///  @return Derived* in variable d;
    ///
    ///  @warning -- this may not work correctly in complex class
    ///              hierarchies with multiple ways of getting from
    ///              derived to base -- particularly if virtual public
    ///              base classes are involved.
    ///
    ///  This function exists for the purpose of writing code in a base
    ///  class that can do things to derived classes - so long as those
    ///  derived classes have well known interfaces.  Basically, it 
    ///  replaces virtual methods with non-virtual methods --
    ///  in an extremely limited set of circumstances.
    ///
    ///  It is normally easy to get the base class pointer in some
    ///  method of a derived class, but difficult without this function
    ///  to get to a derived class.
{
    static Derived *d= (Derived *)(0x100);  // can't use 0 here.
    static Base    *b=d;  // if Derived is not derived from Base this will cause a compile error.

    static size_t  const  bOffset = ((char *)b) - ((char *)d);

    dref =  (Derived*)( ((char*)bptr) - bOffset);

}

template<class T, class U, class W=false_value_type, class X=false_value_type>
struct EvalOr
   ///  Assuming that the input template parameter types are designed to contain
   ///  a compile time constant member named value which evaluates to either true or false,
   ///  construct a compile time constant in this class, also named value, which 
   ///  will be true if any of the template parameter classes is true.
{
   enum { value = T::value || U::value || W::value || X::value };
};


template<class T, class U, class W=true_value_type, class X=true_value_type>
struct EvalAnd
   ///  Assuming that the input template parameter types are designed to contain
   ///  a compile time constant member, named value, which evaluates to either true or false,
   ///  construct a compile time constant in this class, also named value, which 
   ///  will be true if and only if all the template parameter classes are true.
{
   enum { value = T::value && U::value && W::value && X::value};
};



template<class T>
struct  removeAllExtents
   /// Like removeCV<T>, this function removes attributes from its template
   /// parameter T.  In this case, it removes arrayness, constness, pointerness,
   /// and volatility.  It defines a member type, type, which is the same
   /// as it's template parameter, T, but without *, &, [], or volatile keywords.
   /// In sort, it tells you what you are really looking at.
{

   typedef typename removeReference<T>::type                 ReferenceStripped;
   typedef typename pointerReferent<ReferenceStripped>::type PointerStripped;
   typedef typename removeCV<T>::type                        CVStripped;

   typedef CVStripped type;
    
};

template<class T, unsigned long long N>
struct  removeAllExtents<T [N]>
   /// See removeAllExtents<T>
{

  typedef typename removeAllExtents<T>::type type;
    
};



} // namespace cxxtls

#endif
