#include <algorithm>
#include <vector>


namespace cxxtls
{

// note:  used without permission -- seek the copyright from the sources enumerated below.

// See http://www-igm.univ-mlv.fr/~lecroq/seqcomp/node4.html.
// https://en.wikibooks.org/w/index.php?title=Algorithm_Implementation/Strings/Longest_common_subsequence

class LongestCommonSubsequence
    ///  A wrapper around functions that find and print the longest common subsequence between two
    ///  sequences.  See printDiffs() -- but this is not the final output format of the diff unix
    ///  program.  Instead, is raw low level differencing format -- line by line.  
{
    class LCSTable
    {
        size_t   m_;
        size_t   n_;
        size_t*  data_;
    public:
        LCSTable(size_t m, size_t n)
        : m_(m)
        , n_(n)
        {
            data_ = new size_t[(m_ + 1) * (n_ + 1)];
        }
        ~LCSTable()
        {
            delete [] data_;
        }

        void setAt(size_t i, size_t j, size_t value)
        {
            data_[i + j * (m_ + 1)] = value;
        }

        size_t getAt(size_t i, size_t j) const
        {
            return data_[i + j * (m_ + 1)];
        }

        template<typename T> void
        build(const T* X, const T* Y)
        {
            for (size_t i=0; i<=m_; ++i)
                setAt(i, 0, 0);

            for (size_t j=0; j<=n_; ++j)
                setAt(0, j, 0);

            for (size_t i = 0; i < m_; ++i)
            {
                for (size_t j = 0; j < n_; ++j)
                {
                    if (X[i] == Y[j])
                        setAt(i+1, j+1, getAt(i, j)+1);
                    else
                    {
                        size_t left  = getAt(i+1, j);
                        size_t right = getAt(i, j+1);

#ifdef max
                        size_t maximum = max( left, right );
#else
                        size_t maximum = std::max( left, right );
#endif

                        setAt(i+1, 
                              j+1, 
                              maximum
                             );
                    }
                }
            }
        }
    };

    template<typename T, typename accumulator>
    static
    void accumulateDiffs(accumulator &accum, 
                   const LCSTable& C, 
                   T  *X, 
                   T  *Y, 
                   size_t i, 
                   size_t j
                  )
        /// accumulate differences at an item by item into a
        /// container object.  acumulated objects have 3 parts:
        ///
        ///  @li a type flag (' ', '>', or '<')
        ///
        ///  @li the row number (0 based) into the X data array
        ///  @li the row number (0 based) into the Y data array
        ///
        /// The accumulator can print the accumulated information or
        /// aggregate it and print that instead.
    {
        if (i > 0 && j > 0 && X[i-1] == Y[j-1])
        {
            accumulateDiffs(accum, C, X, Y, i-1, j-1);
    
            accum(' ', X[i-1], i, j);
        }
        else
        {
            if (j > 0 && (i == 0 || C.getAt(i,j-1) >= C.getAt(i-1,j)) )
            {
                accumulateDiffs(accum, C, X, Y, i, j-1);
            
                accum('>', Y[j-1], i, j);
    
            }
            else if( i > 0 && (j == 0 || C.getAt(i,j-1) < C.getAt(i-1,j)))
            {
                accumulateDiffs(accum, C, X, Y, i-1, j);
    
                accum('<',X[i-1], i, j);
            }
        }
    }


    template<typename T> static void
    backtrackOne(const LCSTable& table,
                 const T* X, const T* Y, size_t i, size_t j,
                 std::vector<T>& result)
    {
        // used in finding the longest common subsequence

        result.clear();
        if (i == 0 || j == 0)
            return;
        if (X[i - 1] == Y[j - 1])
        {
            backtrackOne(table, X, Y, i - 1, j - 1, result);
            result.push_back(X[i - 1]);
            return;
        }
        if (table.getAt(i, j - 1) > table.getAt(i -1, j))
            backtrackOne(table, X, Y, i, j - 1, result);
        else
            backtrackOne(table, X, Y, i - 1, j, result);
    }

public:
    template<typename T> 
    static void
    findLongestSequence(T* X, 
                        size_t m, 
                        T* Y, 
                        size_t n,
                        std::vector<T>& result
                       )
       // fetch the longest common subsequence -- note -- it will not reflect a 
       // contiguous sequence of items from either of the two input sequences,
       // but rather it will the longest discernable thread of common items found
       // in both.  See wikipedia for a better definition of the longest common
       // subsequence.
    {
        LCSTable table(m, n);
        table.build(X, Y);
        backtrackOne(table, X, Y, m, n, result);
    }

    template<typename T, typename accumulator> 
    static void
    accumulateRawDiffs(
               T      *X, 
               size_t       xLength,
               T      *Y, 
               size_t       yLength,
               accumulator &accum
             )
    /// print (well actually "accumulate") diffs into a container object
    /// at the individual item level -- one item  at a time being compared
    /// to the corresponding item in the second sequence.  X and Y are the
    /// sequences.
    ///
    /// The accumulator object passed in can print the raw data to a stream
    /// or it can aggregate it and print only top level information, or whatever.
    ///
    {

        LCSTable table(xLength, yLength);

        table.build(X, Y);

        accumulateDiffs(accum, 
                  table,
                  X, 
                  Y, 
                  xLength,
                  yLength
                  );
    }


    template<typename stream, typename T>
    struct RawStreamDiffAccumulator
       ///  An object that can accumulate differences of various types into some useful
       ///  container.  In this particular instance the "container" is a std::ostream
       ///  object and the accumulation is merely the act of printing the information
       ///  to the stream.
       ///
       ///  However, a different implementation of this interface can be used to produce
       ///  a normal "diff" style of output by keeping track of the row numbers from the
       ///  two input sequences and showing only the deletes and inserts into stead of 
       ///  all the data.

    {
        stream &outstream_;

        RawStreamDiffAccumulator(stream &s)
        :  outstream_(s)
        {
        }

        void operator() (char     type,  // minus, plus, or blank.  blank means not changed.  plus means deleted
                         T const &item,
                         size_t   i,    // first sequence array index
                         size_t   j     // second sequence array index
                        )
        {
            outstream_ << type << ' ' << item << '\n';
        }
         

    };

    template<typename stream, typename T>
    struct CookedStreamDiffAccumulator
        /// a diff accumulator that works like unix diff command does -- kinda
    {
        stream &outstream_;

        size_t xStartRow_;
        size_t yStartRow_;

        std::vector<T> buffer_;

        char type_;


        CookedStreamDiffAccumulator(stream &s)
        :  outstream_(s)
        ,  xStartRow_(0)
        ,  yStartRow_(0)
        ,  type_(' ')
        {
        }

        void flushBuffer(size_t xRow, size_t yRow)
        {
            char const *direction = "    ";


            if(type_ == '>')
            {
               direction = "add ";
            }
            else
            {
               direction = "del ";
            }

            outstream_ << direction << "@" << xStartRow_ << "/" << yStartRow_ << '\n';



            for(size_t i = 0; i < buffer_.size(); ++i)
            {
               outstream_ << type_ << '\t' << buffer_[i] << "\n";
            }

            buffer_.clear();

            xStartRow_ = xRow;
            yStartRow_ = yRow;

            type_ = ' ';
        }


        void operator() (char     type,  // minus, plus, or blank.  blank means not changed.  plus means deleted
                         T const &item,
                         size_t   i,    // first sequence array index
                         size_t   j     // second sequence array index
                        )
        {
            switch(type)
            {
               case ' ':
                   flushBuffer(i,j);
                   break;
                   
               case '>':
               case '<':

                   if(type != type_)
                   {
                      if(type_ != ' ')
                         flushBuffer(i,j);
                   }

                   type_ = type;
                   
                   buffer_.push_back(item);

                   break;
            }
        }
         

    };



    template<typename T, typename stream> static void
    printDiffs(
                    stream &outstream, 
                    T      *X, 
                    size_t m,   // length of X
                    T* Y, 
                    size_t n    // length of Y
                   )
    {
        RawStreamDiffAccumulator<stream,T> accum(outstream);

        accumulateRawDiffs(X, m, Y, n, accum);
    }


};

};
