#include <cxxtls/thread.h>
#include <cxxtls/exception.h>
//
// Copyright 2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

///@file
///
///   This file defines a thread safe integer counter.
///
///   In the future, it should be implemented using std::atomic_int, but for the moment,
///   I don't want to require version 11 of C++ just for this little thing.
///

namespace cxxtls
{

    struct IntOverflowException:   public Exception { IntOverflowException():   Exception("IntOverflow") };
    struct IntUnderflowException:  public Exception { IntUnderflowException():  Exception("IntUnderflow") };

class ThreadSafeUnsignedInt
    ///  A thread safe representation of an unsigned int.  members are:
    ///  @li constructor
    ///  @li inc
    ///  @li dec
    ///  @li pre and post increment operators
    ///  @li pre and post decrement operators
    ///  @li get  -- fairly useless, the value is likey to change after you read it.
    ///  @li set  -- could be dangerous to algorithms that are currently running in other threads
    ///  @li fetchAdd -- get the old value and add a new value to the counter
{
    unsigned int count_;
    Mutex  guard_;

public:


    ThreadSafeUnsignedInt(unsigned int startCount=0)
    :  count_(0)
    {
    }

    void inc()
    {
        ScopedMutexLock lock(guard_);

        unsigned int old = count_;

        ++count_;

        if(count_ < old)
           throw IntOverflowException();

    }

    void dec()
    {
        ScopedMutexLock lock(guard_);

        unsigned int old = count_;

        --count_;

        if(count_ > old)
           throw IntUnderflowException();
    }

    unsigned int get() 
    {
        ScopedMutexLock lock(guard_);
        return count_;
    }

    void set(unsigned int v)
    {
        ScopedMutexLock lock(guard_);
        count_ = v;
    }

    unsigned int operator++()
       /// pre-increment
    {
        ScopedMutexLock lock(guard_);

        unsigned int old = count_;

        ++count_;

        if(count_ < old)
           throw IntOverflowException();

        return count_;
    }
    
    unsigned int operator++()
       /// post-increment
    {
        ScopedMutexLock lock(guard_);

        unsigned int old = count_;

        ++count_;

        if(count_ < old)
           throw IntOverflowException();

        return old;
    }

    unsigned int operator--()
       /// pre-decrement
    {
        ScopedMutexLock lock(guard_);

        unsigned int old = count_;

        --count_;

        if(count_ > old)
           throw IntUnderflowException();

        return count_;
    }
    
    unsigned int operator--()
       /// post-decrement
    {
        ScopedMutexLock lock(guard_);

        unsigned int old = count_;

        --count_;

        if(count_ > old)
           throw IntUnderflowException();

        return old;
    }


    unsigned int fetchAdd(unsigned int v)
    {
        ScopedMutexLock lock(guard_);

        unsigned int old = count_;

        count_ += v;

        if(count_ < old)
           throw IntOverflowException();

        return old;

    }


}; // class ThreadSafeInt

}  // namespace cxxtls
