#ifdef __linux__
//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

///@file
///  This file implements class cxxtls::PseudoTerminal

#include <cxxtls/pseudoterminal.h>
#include <cxxtls/terminalstatus.h>

#include <term.h>
#include <termios.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <memory>
#include <string.h>

//==============================================================================================
//==============================================================================================
//==============================================================================================
namespace {

cxxtls::TerminalStatus stdinStatus;  // constructor gets terminal status during pre-main static init

char const *stateName[]=
{
        "cxxtls::PseudoTerminal::good",
        "cxxtls::PseudoTerminal::unopenable",
        "cxxtls::PseudoTerminal::unattributable",
        "cxxtls::PseudoTerminal::ungrantable",
        "cxxtls::PseudoTerminal::locked",
        "cxxtls::PseudoTerminal::unnameable",
        "cxxtls::PseudoTerminal::unopened",
};

size_t stateNames = sizeof(stateName) / sizeof(char*);

}

//==============================================================================================
//==============================================================================================
//==============================================================================================

namespace cxxtls
{
//==============================================================================================
//==============================================================================================
//==============================================================================================
void
PseudoTerminal::
setWindowSize(int rows, 
              int cols)
{
    struct winsize size;

    memset(&size, 0, sizeof(size));

    size.ws_row = rows;
    size.ws_col = cols;

    ioctl(masterFD_, TIOCSWINSZ, &size);
     
}

//==============================================================================================
//==============================================================================================
//==============================================================================================

PseudoTerminal::PseudoTerminal()
{
    masterFD_          = -1;
    slaveTerminalName_ = "";
    state_             = unopenable;
    raw_               = false;
}
//==============================================================================================
//==============================================================================================
//==============================================================================================

PseudoTerminal::~PseudoTerminal()
{
    close();
}

//==============================================================================================
//==============================================================================================
//==============================================================================================

// before main runs, get the stdin terminal status

PseudoTerminal::StateValue 
PseudoTerminal::
open(bool raw)
{
     raw_ = raw;

     ScopedMutexLock lock(mutex_);

     int options = O_RDWR;

     masterFD_ = posix_openpt(options);

     if(masterFD_ < 0)
     {
        return unopenable;
     }

     // During program static initialization, the stdinStatus is read.
     // It contains the terminal characteristics that were in effect when
     // the program was first started.
     //
     // We are going to use the default terminal characteristics to
     // initialize the pty's attibutes.  But we are also going to turn
     // on the raw i/o mode feature.

     if(raw_)
         stdinStatus.setRaw();  // make the next call, to setStatus, use raw i/o mode
                                // for the terminal

     if(  stdinStatus.setStatus(masterFD_) < 0)
     {
         ::close(masterFD_);  masterFD_ = -1;
         return unattributable;
         
     }

     if(grantpt(masterFD_) < 0)
     {
         ::close(masterFD_);  masterFD_ = -1;
         return ungrantable;
     }

     if(unlockpt(masterFD_) < 0)
     {
        ::close(masterFD_);  masterFD_ = -1;
        return locked;
     }

     char const *slavePtsNamePtr = ptsname(masterFD_);

     if(!slavePtsNamePtr)
     {
         ::close(masterFD_);  masterFD_ = -1;
         return unnameable;
     }

     slaveTerminalName_ = slavePtsNamePtr;

     state_ = good;

#if 0

     WHOA!  THIS REALLY MESSES UP GdbController

     // guarantee at least one open/close sequence on the slave -- otherwise
     // when you close the masterFD(), with no input ever having taken place
     // any function, in another thread, reading from the masterFD, will NOT
     // receive end of file notification.  I'm guessing this is a linux bug
     // but it is harmless to do everywhere...

     {
        int fh = ::open(slavePtsNamePtr, O_RDWR);

        ::close(fh);

     }
#endif


     return state_;

} // open()

//==============================================================================================
//==============================================================================================
//==============================================================================================

std::string PseudoTerminal::stateNameString() const 
{ 
    ScopedMutexLock lock(  const_cast<Mutex&>(mutex_) );

    if( size_t(state_) >= stateNames )
    {
        return "unknown";
    }

    return stateName[state_]; 
}

//==============================================================================================
//==============================================================================================
//==============================================================================================

void 
PseudoTerminal::
close()
{
    ScopedMutexLock lock(mutex_);

    if(masterFD_ < 0)
        return; // no open

    ::close(masterFD_);
    masterFD_ = -1;
    state_ = unopened;
    slaveTerminalName_ = "";
}

//==============================================================================================
//==============================================================================================
//==============================================================================================

} // namespace cxxtls
//==============================================================================================
//==============================================================================================
//==============================================================================================

#endif
